@@ -17,7 +17,12 @@
 #include "cdi.h"
 #include "cdf_config.h"
 
-/* dummy use of unused parameters to silence compiler warnings */
+// Base file types
+
+#define  CDI_FILETYPE_GRIB        100   // File type GRIB
+#define  CDI_FILETYPE_NETCDF      101   // File type NetCDF
+
+// dummy use of unused parameters to silence compiler warnings
 #ifndef  UNUSED
 #define  UNUSED(x) (void)x
 #endif
@@ -42,7 +47,7 @@ char *strToLower(char *str);
 bool strStartsWith(const char *vstr, const char *cstr);
 
 static inline bool
-strIsEqual(const char *x, const char *y) 
+strIsEqual(const char *x, const char *y)
 {
   return (*x == *y) && strcmp(x, y) == 0;
 }
@@ -238,8 +243,10 @@ VCT;
 enum {
   CDF_DIMID_X,
   CDF_DIMID_Y,
+  CDF_DIMID_RP, // reducedPoints
   CDF_VARID_X,
   CDF_VARID_Y,
+  CDF_VARID_RP, // reducedPoints
   CDF_VARID_A,
   CDF_SIZE_ncIDs,
 };
@@ -275,12 +282,13 @@ typedef struct {
   int         ncmode;
   int         vlistID;
 #ifdef HAVE_LIBNETCDF
+  int         nc_complex_float_id;
+  int         nc_complex_double_id;
   ncgrid_t    ncgrid[MAX_GRIDS_PS];
   int         zaxisID[MAX_ZAXES_PS];	// Warning: synchronous array to vlist_to_pointer(vlistID)->zaxisIDs
   int         nczvarID[MAX_ZAXES_PS];
   VCT         vct;
 #endif
-  int         historyID;
   int         globalatts;
   int         localatts;
   int         unreduced;
@@ -510,6 +518,7 @@ void cdi_check_gridsize_int_limit(const char *format, size_t gridsize)
   if ( gridsize > INT_MAX ) Error("%s format grid size (%zu) limit exceeded (%zu)!", format, gridsize, INT_MAX);
 }
 
+int cdiBaseFiletype(int filetype);
 
 #if defined (__cplusplus)
 }
