@@ -2,7 +2,7 @@
   This file is part of CDO. CDO is a collection of Operators to
   manipulate and analyse Climate model Data.
 
-  Copyright (C) 2003-2019 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
+  Copyright (C) 2003-2020 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
   See COPYING file for copying and redistribution conditions.
 
   This program is free software; you can redistribute it and/or modify
@@ -26,9 +26,10 @@
 */
 
 #include <cdi.h>
-#include "cdo_options.h"
 
-#include "cdo_int.h"
+#include "cdo_options.h"
+#include "functs.h"
+#include "process_int.h"
 #include "calendar.h"
 
 static double
@@ -43,12 +44,12 @@ dayofyear(int calendar, int64_t vdate, int vtime)
   cdiDecodeDate(vdate, &year, &month, &day);
   cdiDecodeTime(vtime, &hour, &minute, &second);
 
-  const int dpy = days_per_year(calendar, year);
+  const auto dpy = days_per_year(calendar, year);
 
   double doy = 0;
   for (int im = 1; im < month; ++im)
     {
-      const int *dpm = (dpy == 360) ? month_360 : ((dpy == 365) ? month_365 : month_366);
+      auto dpm = (dpy == 360) ? month_360 : ((dpy == 365) ? month_365 : month_366);
       if (im >= 1 && im <= 12) doy += dpm[im - 1];
     }
 
@@ -77,25 +78,27 @@ Arithdays(void *process)
   int MULDOY = cdoOperatorAdd("muldoy", func_mul,         0,  nullptr);
   // clang-format on
 
-  const int operatorID = cdoOperatorID();
-  const int operfunc = cdoOperatorF1(operatorID);
-  const int operfunc2 = cdoOperatorF2(operatorID);
+  const auto operatorID = cdoOperatorID();
+  const auto operfunc = cdoOperatorF1(operatorID);
+  const auto operfunc2 = cdoOperatorF2(operatorID);
 
-  CdoStreamID streamID1 = cdoOpenRead(0);
+  operatorCheckArgc(0);
 
-  const int vlistID1 = cdoStreamInqVlist(streamID1);
-  const int vlistID2 = vlistDuplicate(vlistID1);
+  const auto streamID1 = cdoOpenRead(0);
 
-  const int taxisID1 = vlistInqTaxis(vlistID1);
-  const int taxisID2 = taxisDuplicate(taxisID1);
+  const auto vlistID1 = cdoStreamInqVlist(streamID1);
+  const auto vlistID2 = vlistDuplicate(vlistID1);
+
+  const auto taxisID1 = vlistInqTaxis(vlistID1);
+  const auto taxisID2 = taxisDuplicate(taxisID1);
   vlistDefTaxis(vlistID2, taxisID2);
 
-  const int calendar = taxisInqCalendar(taxisID1);
+  const auto calendar = taxisInqCalendar(taxisID1);
 
-  CdoStreamID streamID2 = cdoOpenWrite(1);
+  const auto streamID2 = cdoOpenWrite(1);
   cdoDefVlist(streamID2, vlistID2);
 
-  const size_t gridsizemax = vlistGridsizeMax(vlistID1);
+  const auto gridsizemax = vlistGridsizeMax(vlistID1);
 
   Field field;
   field.resize(gridsizemax);
@@ -103,8 +106,8 @@ Arithdays(void *process)
   int tsID = 0;
   while ((nrecs = cdoStreamInqTimestep(streamID1, tsID)))
     {
-      const int64_t vdate = taxisInqVdate(taxisID1);
-      const int vtime = taxisInqVtime(taxisID1);
+      const auto vdate = taxisInqVdate(taxisID1);
+      const auto vtime = taxisInqVtime(taxisID1);
 
       taxisCopyTimestep(taxisID2, taxisID1);
       cdoDefTimestep(streamID2, tsID);
@@ -122,7 +125,7 @@ Arithdays(void *process)
       for (int recID = 0; recID < nrecs; recID++)
         {
           cdoInqRecord(streamID1, &varID, &levelID);
-          cdoReadRecord(streamID1, field.vec.data(), &field.nmiss);
+          cdoReadRecord(streamID1, field.vec_d.data(), &field.nmiss);
 
           field.grid = vlistInqVarGrid(vlistID1, varID);
           field.missval = vlistInqVarMissval(vlistID1, varID);
@@ -130,7 +133,7 @@ Arithdays(void *process)
           vfarcfun(field, rconst, operfunc);
 
           cdoDefRecord(streamID2, varID, levelID);
-          cdoWriteRecord(streamID2, field.vec.data(), field.nmiss);
+          cdoWriteRecord(streamID2, field.vec_d.data(), field.nmiss);
         }
 
       tsID++;
