@@ -2,7 +2,7 @@
   This file is part of CDO. CDO is a collection of Operators to
   manipulate and analyse Climate model Data.
 
-  Copyright (C) 2003-2019 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
+  Copyright (C) 2003-2020 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
   See COPYING file for copying and redistribution conditions.
 
   This program is free software; you can redistribute it and/or modify
@@ -33,39 +33,42 @@
 
 #include <cdi.h>
 
-#include "cdo_int.h"
+#include "cdo_options.h"
+#include "functs.h"
+#include "process_int.h"
 #include "param_conversion.h"
 #include <mpim_grid.h>
-#include "percentiles.h"
 #include "pmlist.h"
+#include "cdo_zaxis.h"
 
+template <typename T>
 static void
-print_location_LL(int operfunc, int vlistID, int varID, int levelID, int gridID, double sglval,
-                  const std::vector<double> &fieldvec, int64_t vdate, int vtime)
+print_location_LL(int operfunc, int vlistID, int varID, int levelID, int gridID, T sglval, const Varray<T> &fieldvec,
+                  int64_t vdate, int vtime)
 {
-  static bool showHeader = true;
-  const int code = vlistInqVarCode(vlistID, varID);
+  static auto showHeader = true;
+  const auto code = vlistInqVarCode(vlistID, varID);
 
   int year, month, day, hour, minute, second;
   cdiDecodeDate(vdate, &year, &month, &day);
   cdiDecodeTime(vtime, &hour, &minute, &second);
 
-  const bool lreg2d = (gridInqType(gridID) == GRID_GAUSSIAN || gridInqType(gridID) == GRID_LONLAT);
+  const auto lreg2d = (gridInqType(gridID) == GRID_GAUSSIAN || gridInqType(gridID) == GRID_LONLAT);
 
   if (lreg2d || gridInqType(gridID) == GRID_CURVILINEAR || gridInqType(gridID) == GRID_UNSTRUCTURED)
     {
-      const int zaxisID = vlistInqVarZaxis(vlistID, varID);
-      const double level = cdoZaxisInqLevel(zaxisID, levelID);
-      const size_t gridsize = gridInqSize(gridID);
-      const size_t nlon = gridInqXsize(gridID);
+      const auto zaxisID = vlistInqVarZaxis(vlistID, varID);
+      const auto level = cdoZaxisInqLevel(zaxisID, levelID);
+      const auto gridsize = gridInqSize(gridID);
+      const auto nlon = gridInqXsize(gridID);
       for (size_t ij = 0; ij < gridsize; ++ij)
         {
           if (DBL_IS_EQUAL(fieldvec[ij], sglval))
             {
-              const size_t j = ij / nlon;
-              const size_t i = ij - j * nlon;
-              const double xval = gridInqXval(gridID, lreg2d ? i : ij);
-              const double yval = gridInqYval(gridID, lreg2d ? j : ij);
+              const auto j = ij / nlon;
+              const auto i = ij - j * nlon;
+              const auto xval = gridInqXval(gridID, lreg2d ? i : ij);
+              const auto yval = gridInqYval(gridID, lreg2d ? j : ij);
               if (showHeader)
                 {
                   fprintf(stdout, "  Date     Time     Code  Level   Lon      Lat          %s\n",
@@ -84,10 +87,10 @@ print_location_LL(int operfunc, int vlistID, int varID, int levelID, int gridID,
 static void
 fldstatGetParameter(bool *weights)
 {
-  const int pargc = operatorArgc();
+  const auto pargc = operatorArgc();
   if (pargc)
     {
-      const char **pargv = (const char **) operatorArgv();
+      auto pargv = cdoGetOperArgv();
 
       KVList kvlist;
       kvlist.name = "FLDSTAT";
@@ -96,15 +99,15 @@ fldstatGetParameter(bool *weights)
 
       for (const auto &kv : kvlist)
         {
-          const char *key = kv.key.c_str();
-          if (kv.nvalues > 1) cdoAbort("Too many values for parameter key >%s<!", key);
-          if (kv.nvalues < 1) cdoAbort("Missing value for parameter key >%s<!", key);
-          const char *value = kv.values[0].c_str();
+          const auto &key = kv.key;
+          if (kv.nvalues > 1) cdoAbort("Too many values for parameter key >%s<!", key.c_str());
+          if (kv.nvalues < 1) cdoAbort("Missing value for parameter key >%s<!", key.c_str());
+          const auto &value = kv.values[0];
 
-          if (cstrIsEqual(key, "weights"))
+          if (key == "weights")
             *weights = parameter2bool(value);
           else
-            cdoAbort("Invalid parameter key >%s<!", key);
+            cdoAbort("Invalid parameter key >%s<!", key.c_str());
         }
     }
 }
@@ -123,9 +126,9 @@ addOperators(void)
   cdoOperatorAdd("fldstd1",  func_std1w,  1, nullptr);
   cdoOperatorAdd("fldvar",   func_varw,   1, nullptr);
   cdoOperatorAdd("fldvar1",  func_var1w,  1, nullptr);
-  cdoOperatorAdd("fldpctl",  func_pctl,   0, nullptr);
   cdoOperatorAdd("fldskew",  func_skew,   0, nullptr);
   cdoOperatorAdd("fldkurt",  func_kurt,   0, nullptr);
+  cdoOperatorAdd("fldpctl",  func_pctl,   0, nullptr);
   // clang-format on
 }
 
@@ -140,50 +143,51 @@ Fldstat(void *process)
 
   addOperators();
 
-  const int operatorID = cdoOperatorID();
-  const int operfunc = cdoOperatorF1(operatorID);
+  const auto operatorID = cdoOperatorID();
+  const auto operfunc = cdoOperatorF1(operatorID);
   const bool needWeights = cdoOperatorF2(operatorID) != 0;
 
   double pn = 0;
   if (operfunc == func_pctl)
     {
       operatorInputArg("percentile number");
-      pn = parameter2double(operatorArgv()[0]);
-      percentile_check_number(pn);
+      pn = parameter2double(cdoOperatorArgv(0));
+    }
+  else
+    {
+      operatorCheckArgc(0);
     }
 
   bool useweights = true;
   if (needWeights) fldstatGetParameter(&useweights);
 
-  CdoStreamID streamID1 = cdoOpenRead(0);
+  const auto streamID1 = cdoOpenRead(0);
 
-  const int vlistID1 = cdoStreamInqVlist(streamID1);
-  const int vlistID2 = vlistDuplicate(vlistID1);
+  const auto vlistID1 = cdoStreamInqVlist(streamID1);
+  const auto vlistID2 = vlistDuplicate(vlistID1);
 
-  const int taxisID1 = vlistInqTaxis(vlistID1);
-  const int taxisID2 = taxisDuplicate(taxisID1);
+  const auto taxisID1 = vlistInqTaxis(vlistID1);
+  const auto taxisID2 = taxisDuplicate(taxisID1);
   vlistDefTaxis(vlistID2, taxisID2);
 
-  const double slon = 0;
-  const double slat = 0;
-  const int gridID2 = gridCreate(GRID_LONLAT, 1);
+  const auto slon = 0.0, slat = 0.0;
+  const auto gridID2 = gridCreate(GRID_LONLAT, 1);
   gridDefXsize(gridID2, 1);
   gridDefYsize(gridID2, 1);
   gridDefXvals(gridID2, &slon);
   gridDefYvals(gridID2, &slat);
 
-  const int ngrids = vlistNgrids(vlistID1);
+  const auto ngrids = vlistNgrids(vlistID1);
 
   for (int index = 0; index < ngrids; index++) vlistChangeGridIndex(vlistID2, index, gridID2);
 
-  CdoStreamID streamID2 = cdoOpenWrite(1);
+  const auto streamID2 = cdoOpenWrite(1);
 
   cdoDefVlist(streamID2, vlistID2);
 
-  const size_t gridsizemax = vlistGridsizeMax(vlistID1);
+  const auto gridsizemax = vlistGridsizeMax(vlistID1);
 
   Field field;
-  field.resize(gridsizemax);
   if (needWeights)
     {
       field.weightv.resize(gridsizemax);
@@ -194,23 +198,23 @@ Fldstat(void *process)
         }
     }
 
+  VarList varList1;
+  varListInit(varList1, vlistID1);
+
   int tsID = 0;
   while ((nrecs = cdoStreamInqTimestep(streamID1, tsID)))
     {
       taxisCopyTimestep(taxisID2, taxisID1);
       cdoDefTimestep(streamID2, tsID);
 
-      // Precompute date + time for later representation in verbose mode
-      const int64_t vdate = (Options::cdoVerbose && (operfunc == func_min || operfunc == func_max)) ? taxisInqVdate(taxisID1) : 0;
-      const int vtime = (Options::cdoVerbose && (operfunc == func_min || operfunc == func_max)) ? taxisInqVtime(taxisID1) : 0;
+      const auto vdate = taxisInqVdate(taxisID1);
+      const auto vtime = taxisInqVtime(taxisID1);
 
       for (int recID = 0; recID < nrecs; recID++)
         {
           cdoInqRecord(streamID1, &varID, &levelID);
-          cdoReadRecord(streamID1, field.vec.data(), &field.nmiss);
-
-          field.grid = vlistInqVarGrid(vlistID1, varID);
-          field.size = gridInqSize(field.grid);
+          field.init(varList1[varID]);
+          cdoReadRecord(streamID1, field);
 
           if (needWeights && field.grid != lastgrid)
             {
@@ -222,30 +226,28 @@ Fldstat(void *process)
                   if (wstatus && tsID == 0 && levelID == 0)
                     {
                       if (ngrids == 1)
-                        {
-                          cdoWarning("Grid cell bounds not available, using constant grid cell area weights!");
-                        }
+                        cdoWarning("Grid cell bounds not available, using constant grid cell area weights!");
                       else
-                        {
-                          char varname[CDI_MAX_NAME];
-                          vlistInqVarName(vlistID1, varID, varname);
-                          cdoWarning("Grid cell bounds not available, using constant grid cell area weights for variable %s!",
-                                     varname);
-                        }
+                        cdoWarning("Grid cell bounds not available, using constant grid cell area weights for variable %s!",
+                                   varList1[varID].name);
                     }
                 }
             }
 
-          field.missval = vlistInqVarMissval(vlistID1, varID);
-          double sglval = (operfunc == func_pctl) ? vfldpctl(field, pn) : vfldfun(field, operfunc);
+          auto singleValue = (operfunc == func_pctl) ? fieldPctl(field, pn) : fieldFunction(field, operfunc);
 
           if (Options::cdoVerbose && (operfunc == func_min || operfunc == func_max))
-            print_location_LL(operfunc, vlistID1, varID, levelID, field.grid, sglval, field.vec, vdate, vtime);
+            {
+              if (field.memType == MemType::Float)
+                print_location_LL(operfunc, vlistID1, varID, levelID, field.grid, (float)singleValue, field.vec_f, vdate, vtime);
+              else
+                print_location_LL(operfunc, vlistID1, varID, levelID, field.grid, singleValue, field.vec_d, vdate, vtime);
+            }
 
-          const size_t nmiss = DBL_IS_EQUAL(sglval, field.missval);
+          const size_t nmiss = DBL_IS_EQUAL(singleValue, field.missval);
 
           cdoDefRecord(streamID2, varID, levelID);
-          cdoWriteRecord(streamID2, &sglval, nmiss);
+          cdoWriteRecord(streamID2, &singleValue, nmiss);
         }
 
       tsID++;
