@@ -27,10 +27,12 @@
 
 #include <cdi.h>
 
-#include "cdo_int.h"
+#include "util_date.h"
+#include "functs.h"
+#include "process_int.h"
+#include "cdo_vlist.h"
 #include "param_conversion.h"
 #include "percentiles_hist.h"
-#include "percentiles.h"
 #include "datetime.h"
 
 static void
@@ -39,66 +41,65 @@ timpctl(int operatorID)
   TimeStat timestat_date = TimeStat::MEAN;
   char indate1[DATE_LEN + 1], indate2[DATE_LEN + 1];
   int nrecs;
-  int gridID, varID, levelID;
   size_t nmiss;
-  int nlevels;
 
   operatorInputArg("percentile number");
-  const double pn = parameter2double(operatorArgv()[0]);
-  percentile_check_number(pn);
+  const auto pn = parameter2double(cdoOperatorArgv(0));
 
   const int cmplen = DATE_LEN - cdoOperatorF2(operatorID);
 
-  CdoStreamID streamID1 = cdoOpenRead(0);
-  CdoStreamID streamID2 = cdoOpenRead(1);
-  CdoStreamID streamID3 = cdoOpenRead(2);
+  const auto streamID1 = cdoOpenRead(0);
+  const auto streamID2 = cdoOpenRead(1);
+  const auto streamID3 = cdoOpenRead(2);
 
-  const int vlistID1 = cdoStreamInqVlist(streamID1);
-  const int vlistID2 = cdoStreamInqVlist(streamID2);
-  const int vlistID3 = cdoStreamInqVlist(streamID3);
-  const int vlistID4 = vlistDuplicate(vlistID1);
+  const auto vlistID1 = cdoStreamInqVlist(streamID1);
+  const auto vlistID2 = cdoStreamInqVlist(streamID2);
+  const auto vlistID3 = cdoStreamInqVlist(streamID3);
+  const auto vlistID4 = vlistDuplicate(vlistID1);
 
   vlistCompare(vlistID1, vlistID2, CMP_ALL);
   vlistCompare(vlistID1, vlistID3, CMP_ALL);
 
   if (cdoOperatorF2(operatorID) == 16) vlistDefNtsteps(vlistID4, 1);
 
-  const int taxisID1 = vlistInqTaxis(vlistID1);
-  const int taxisID2 = vlistInqTaxis(vlistID2);
-  const int taxisID3 = vlistInqTaxis(vlistID3);
+  const auto taxisID1 = vlistInqTaxis(vlistID1);
+  const auto taxisID2 = vlistInqTaxis(vlistID2);
+  const auto taxisID3 = vlistInqTaxis(vlistID3);
   /* TODO - check that time axes 2 and 3 are equal */
 
-  const int taxisID4 = taxisDuplicate(taxisID1);
+  const auto taxisID4 = taxisDuplicate(taxisID1);
   taxisWithBounds(taxisID4);
   vlistDefTaxis(vlistID4, taxisID4);
 
-  CdoStreamID streamID4 = cdoOpenWrite(3);
+  const auto streamID4 = cdoOpenWrite(3);
   cdoDefVlist(streamID4, vlistID4);
 
-  const int nvars = vlistNvars(vlistID1);
+  const auto nvars = vlistNvars(vlistID1);
 
-  const int maxrecs = vlistNrecs(vlistID1);
+  const auto maxrecs = vlistNrecs(vlistID1);
   std::vector<RecordInfo> recList(maxrecs);
 
   DateTimeList dtlist;
   dtlist.setStat(timestat_date);
   dtlist.setCalendar(taxisInqCalendar(taxisID1));
 
-  const size_t gridsizemax = vlistGridsizeMax(vlistID1);
+  const auto gridsizemax = vlistGridsizeMax(vlistID1);
 
   Field field;
   field.resize(gridsizemax);
 
   FieldVector2D vars1;
   fieldsFromVlist(vlistID1, vars1, FIELD_VEC);
-  HISTOGRAM_SET *hset = hsetCreate(nvars);
+  HistogramSet hset(nvars);
 
-  for (varID = 0; varID < nvars; varID++)
+  VarList varList1;
+  varListInit(varList1, vlistID1);
+
+  for (int varID = 0; varID < nvars; varID++)
     {
-      gridID = vlistInqVarGrid(vlistID1, varID);
-      nlevels = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
-
-      hsetCreateVarLevels(hset, varID, nlevels, gridID);
+      const auto gridsize = varList1[varID].gridsize;
+      const auto nlevels = varList1[varID].nlevels;
+      hset.createVarLevels(varID, nlevels, gridsize);
     }
 
   int tsID = 0;
@@ -109,36 +110,38 @@ timpctl(int operatorID)
       if (nrecs != cdoStreamInqTimestep(streamID3, otsID))
         cdoAbort("Number of records at time step %d of %s and %s differ!", otsID + 1, cdoGetStreamName(1), cdoGetStreamName(2));
 
-      const int64_t vdate2 = taxisInqVdate(taxisID2);
-      const int vtime2 = taxisInqVtime(taxisID2);
-      const int64_t vdate3 = taxisInqVdate(taxisID3);
-      const int vtime3 = taxisInqVtime(taxisID3);
+      const auto vdate2 = taxisInqVdate(taxisID2);
+      const auto vtime2 = taxisInqVtime(taxisID2);
+      const auto vdate3 = taxisInqVdate(taxisID3);
+      const auto vtime3 = taxisInqVtime(taxisID3);
       if (vdate2 != vdate3 || vtime2 != vtime3)
         cdoAbort("Verification dates at time step %d of %s and %s differ!", otsID + 1, cdoGetStreamName(1), cdoGetStreamName(2));
 
       for (int recID = 0; recID < nrecs; recID++)
         {
+          int varID, levelID;
           cdoInqRecord(streamID2, &varID, &levelID);
-          cdoReadRecord(streamID2, vars1[varID][levelID].vec.data(), &nmiss);
+          cdoReadRecord(streamID2, vars1[varID][levelID].vec_d.data(), &nmiss);
           vars1[varID][levelID].nmiss = nmiss;
         }
 
       for (int recID = 0; recID < nrecs; recID++)
         {
+          int varID, levelID;
           cdoInqRecord(streamID3, &varID, &levelID);
-          cdoReadRecord(streamID3, field.vec.data(), &field.nmiss);
+          cdoReadRecord(streamID3, field.vec_d.data(), &field.nmiss);
           field.grid = vars1[varID][levelID].grid;
           field.missval = vars1[varID][levelID].missval;
 
-          hsetDefVarLevelBounds(hset, varID, levelID, &vars1[varID][levelID], &field);
+          hset.defVarLevelBounds(varID, levelID, vars1[varID][levelID], field);
         }
 
       int nsets = 0;
       while (nrecs && (nrecs = cdoStreamInqTimestep(streamID1, tsID)))
         {
           dtlist.taxisInqTimestep(taxisID1, nsets);
-          const int64_t vdate1 = dtlist.getVdate(nsets);
-          const int vtime1 = dtlist.getVtime(nsets);
+          const auto vdate1 = dtlist.getVdate(nsets);
+          const auto vtime1 = dtlist.getVtime(nsets);
 
           if (nsets == 0) SET_DATE(indate2, vdate1, vtime1);
           SET_DATE(indate1, vdate1, vtime1);
@@ -147,17 +150,18 @@ timpctl(int operatorID)
 
           for (int recID = 0; recID < nrecs; recID++)
             {
+              int varID, levelID;
               cdoInqRecord(streamID1, &varID, &levelID);
               if (tsID == 0)
                 {
                   recList[recID].varID = varID;
                   recList[recID].levelID = levelID;
-                  recList[recID].lconst = vlistInqVarTimetype(vlistID1, varID) == TIME_CONSTANT;
+                  recList[recID].lconst = varList1[varID].timetype == TIME_CONSTANT;
                 }
-              cdoReadRecord(streamID1, vars1[varID][levelID].vec.data(), &nmiss);
+              cdoReadRecord(streamID1, vars1[varID][levelID].vec_d.data(), &nmiss);
               vars1[varID][levelID].nmiss = nmiss;
 
-              hsetAddVarLevelValues(hset, varID, levelID, &vars1[varID][levelID]);
+              hset.addSubVarLevelValues(varID, levelID, vars1[varID][levelID], 1, FIELD_VEC);
             }
 
           nsets++;
@@ -166,13 +170,12 @@ timpctl(int operatorID)
 
       if (nrecs == 0 && nsets == 0) break;
 
-      for (varID = 0; varID < nvars; varID++)
+      for (int varID = 0; varID < nvars; varID++)
         {
-          if (vlistInqVarTimetype(vlistID1, varID) == TIME_CONSTANT) continue;
-          nlevels = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
+          if (varList1[varID].timetype == TIME_CONSTANT) continue;
 
-          for (levelID = 0; levelID < nlevels; levelID++)
-            hsetGetVarLevelPercentiles(&vars1[varID][levelID], hset, varID, levelID, pn);
+          const auto nlevels = varList1[varID].nlevels;
+          for (int levelID = 0; levelID < nlevels; levelID++) hset.getVarLevelPercentiles(vars1[varID][levelID], varID, levelID, pn, FIELD_VEC);
         }
 
       dtlist.statTaxisDefTimestep(taxisID4, nsets);
@@ -182,18 +185,16 @@ timpctl(int operatorID)
         {
           if (otsID && recList[recID].lconst) continue;
 
-          const int varID = recList[recID].varID;
-          const int levelID = recList[recID].levelID;
+          const auto varID = recList[recID].varID;
+          const auto levelID = recList[recID].levelID;
           cdoDefRecord(streamID4, varID, levelID);
-          cdoWriteRecord(streamID4, vars1[varID][levelID].vec.data(), vars1[varID][levelID].nmiss);
+          cdoWriteRecord(streamID4, vars1[varID][levelID].vec_d.data(), vars1[varID][levelID].nmiss);
         }
 
       if (nrecs == 0) break;
       otsID++;
     }
 
-  hsetDestroy(hset);
-
   cdoStreamClose(streamID4);
   cdoStreamClose(streamID3);
   cdoStreamClose(streamID2);
@@ -213,9 +214,7 @@ Timpctl(void *process)
   cdoOperatorAdd("hourpctl", func_pctl,  4, nullptr);
   // clang-format on
 
-  const int operatorID = cdoOperatorID();
-
-  timpctl(operatorID);
+  timpctl(cdoOperatorID());
 
   cdoFinish();
 
