@@ -2,7 +2,7 @@
   This file is part of CDO. CDO is a collection of Operators to
   manipulate and analyse Climate model Data.
 
-  Copyright (C) 2003-2019 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
+  Copyright (C) 2003-2020 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
   See COPYING file for copying and redistribution conditions.
 
   This program is free software; you can redistribute it and/or modify
@@ -14,8 +14,8 @@
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 */
+
 #include <cmath>
-#include <string.h>
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -23,13 +23,10 @@
 
 #include "cdo_options.h"
 #include "cimdOmp.h"
-#include "dmemory.h"
 #include "constants.h"
-#include "cimdOmp.h"
 #include "mpmo.h"
 #include "array.h"
 
-
 extern "C" void gaussianLatitudes(double *latitudes, double *weights, size_t nlat);
 
 static void
@@ -244,8 +241,8 @@ phcs(double *pnm, double *hnm, long waves, double pmu, double *ztemp1, double *z
   hnm[0] = pmu * ztemp2[0];
 
   for (jn = 1; jn < waves; jn++)
-    hnm[jn]
-        = (jn + 1) * pmu * ztemp2[jn] - std::sqrt(((jn + jn + 3.0) * ((jn + 1) * (jn + 1) - 1.0)) / (jn + jn + 1.0)) * ztemp2[jn - 1];
+    hnm[jn] = (jn + 1) * pmu * ztemp2[jn]
+              - std::sqrt(((jn + jn + 3.0) * ((jn + 1) * (jn + 1) - 1.0)) / (jn + jn + 1.0)) * ztemp2[jn - 1];
 
   hnm += waves;
 
@@ -285,8 +282,8 @@ void
 after_legini_full(long ntr, long nlat, double *restrict poli, double *restrict pold, double *restrict pdev, double *restrict pol2,
                   double *restrict pol3, double *restrict coslat)
 {
-  long waves = ntr + 1;
-  long dimsp = (ntr + 1) * (ntr + 2);
+  long waves = ntr + 1; // used in omp loop
+  long dimsp = (ntr + 1) * (ntr + 2); // used in omp loop
 
   Varray<double> gmu(nlat), gwt(nlat);
   gaussianLatitudes(gmu.data(), gwt.data(), nlat);
@@ -334,8 +331,8 @@ after_legini_full(long ntr, long nlat, double *restrict poli, double *restrict p
 void
 after_legini(long ntr, long nlat, double *restrict poli, double *restrict pold, double *restrict coslat)
 {
-  long waves = ntr + 1;
-  const long dimpnm = (ntr + 1) * (ntr + 4) / 2;
+  long waves = ntr + 1; // used in omp loop
+  long dimpnm = (ntr + 1) * (ntr + 4) / 2; // used in omp loop
 
 #ifdef _OPENMP
   Varray2D<double> pnm2(Threading::ompNumThreads);
@@ -391,53 +388,28 @@ after_legini(long ntr, long nlat, double *restrict poli, double *restrict pold,
     }
 }
 
-/* to slow for nec, 2.0 instead of 2.3 GFlops ( vector length too small ) */
-void
-sp2fctest(const double *sa, double *fa, const double *poli, long nlev, long nlat, long nfc, long nt)
+static inline void
+sp2fc_kernel(long nlat, const double *pol, const double *sal, double *restrict far, double *restrict fai)
 {
-  long nsp2 = (nt + 1) * (nt + 2);
-
-  for (long lev = 0; lev < nlev; lev++)
+  const auto sar = sal[0];
+  const auto sai = sal[1];
+  for (long lat = 0; lat < nlat; lat++)
     {
-      auto pol = poli;
-      auto sal = sa + lev * nsp2;
-      auto fal = fa + lev * nfc * nlat;
-      memset(fal, 0, nfc * nlat * sizeof(double));
-
-      for (long jm = 0; jm <= nt; jm++)
-        {
-          for (long jn = 0; jn <= nt - jm; jn++)
-            {
-              const long is = (jn + 1) % 2 * 2 - 1;
-              const double sar = *sal++;
-              const double sai = *sal++;
-              const double saris = sar * is;
-              const double saiis = sai * is;
-              auto far = fal;
-              auto fai = fal + nlat;
-#ifdef HAVE_OPENMP4
-#pragma omp simd
-#endif
-              for (long latn = 0; latn < nlat / 2; latn++)
-                {
-                  const long lats = nlat - latn - 1;
-                  far[latn] += pol[latn] * sar;
-                  fai[latn] += pol[latn] * sai;
-                  far[lats] += pol[latn] * saris;
-                  fai[lats] += pol[latn] * saiis;
-                }
-              pol += nlat;
-            }
-          fal += 2 * nlat;
-        }
+      far[lat] += pol[lat] * sar;
+      fai[lat] += pol[lat] * sai;
     }
 }
 
 void
 sp2fc(const double *sa, double *fa, const double *poli, long nlev, long nlat, long nfc, long nt)
 {
+  long ntp1 = nt + 1;
   long nsp2 = (nt + 1) * (nt + 2);
 
+  std::vector<long> cumindex(ntp1);
+  cumindex[0] = 0;
+  for (long jmm = 1; jmm < ntp1; jmm++) cumindex[jmm] = cumindex[jmm - 1] + (ntp1 - jmm + 1);
+
 #ifdef _OPENMP
 #pragma omp parallel for default(shared)
 #endif
@@ -448,36 +420,46 @@ sp2fc(const double *sa, double *fa, const double *poli, long nlev, long nlat, lo
       auto fal = fa + lev * nfc * nlat;
       memset(fal, 0, nfc * nlat * sizeof(double));
 
-      for (long jmm = 0; jmm <= nt; jmm++)
+      for (long jmm = 0; jmm < ntp1; jmm++)
         {
-          for (long jfc = jmm; jfc <= nt; jfc++)
+          auto polt = pol + cumindex[jmm] * nlat;
+          auto falt = fal + jmm * 2 * nlat;
+          auto salt = sal + cumindex[jmm] * 2;
+          for (long jfc = 0; jfc < ntp1 - jmm; jfc++)
             {
-              const double sar = *sal++;
-              const double sai = *sal++;
-              double *restrict far = fal;
-              double *restrict fai = fal + nlat;
-              /* unaligned loop start
-#ifdef  HAVE_OPENMP4
-#pragma omp simd
-#endif
-              */
-              for (long lat = 0; lat < nlat; lat++)
-                {
-                  far[lat] += pol[lat] * sar;
-                  fai[lat] += pol[lat] * sai;
-                }
-              pol += nlat;
+              sp2fc_kernel(nlat, polt + jfc * nlat, salt + jfc * 2, falt, falt + nlat);
             }
-          fal += 2 * nlat;
         }
     }
 }
 
+static inline void
+fc2sp_kernel(long nlat, const double *pol, const double *far, const double *fai, double *sal)
+{
+  double sar = 0.0;
+  double sai = 0.0;
+#ifdef HAVE_OPENMP4
+#pragma omp simd reduction(+ : sar) reduction(+ : sai)
+#endif
+  for (long lat = 0; lat < nlat; lat++)
+    {
+      sar += pol[lat] * far[lat];
+      sai += pol[lat] * fai[lat];
+    }
+  sal[0] = sar;
+  sal[1] = sai;
+}
+
 void
 fc2sp(const double *fa, double *sa, const double *poli, long nlev, long nlat, long nfc, long nt)
 {
+  long ntp1 = nt + 1;
   long nsp2 = (nt + 1) * (nt + 2);
 
+  std::vector<long> cumindex(ntp1);
+  cumindex[0] = 0;
+  for (long jmm = 1; jmm < ntp1; jmm++) cumindex[jmm] = cumindex[jmm - 1] + (ntp1 - jmm + 1);
+
 #ifdef _OPENMP
 #pragma omp parallel for default(shared)
 #endif
@@ -487,27 +469,15 @@ fc2sp(const double *fa, double *sa, const double *poli, long nlev, long nlat, lo
       auto fal = fa + lev * nfc * nlat;
       auto sal = sa + lev * nsp2;
 
-      for (long jmm = 0; jmm <= nt; jmm++)
+      for (long jmm = 0; jmm < ntp1; jmm++)
         {
-          for (long jfc = jmm; jfc <= nt; jfc++)
+          auto polt = pol + cumindex[jmm] * nlat;
+          auto falt = fal + jmm * 2 * nlat;
+          auto salt = sal + cumindex[jmm] * 2;
+          for (long jfc = 0; jfc < ntp1 - jmm; jfc++)
             {
-              auto far = fal;
-              auto fai = fal + nlat;
-              double sar = 0.0;
-              double sai = 0.0;
-#ifdef HAVE_OPENMP4
-#pragma omp simd reduction(+ : sar) reduction(+ : sai)
-#endif
-              for (long lat = 0; lat < nlat; lat++)
-                {
-                  sar += pol[lat] * far[lat];
-                  sai += pol[lat] * fai[lat];
-                }
-              *sal++ = sar;
-              *sal++ = sai;
-              pol += nlat;
+              fc2sp_kernel(nlat, polt + jfc * nlat, falt, falt + nlat, salt + jfc * 2);
             }
-          fal += 2 * nlat;
         }
     }
 }
@@ -519,13 +489,11 @@ fc2sp(const double *fa, double *sa, const double *poli, long nlev, long nlat, lo
 void
 sp2sp(double *arrayIn, long truncIn, double *arrayOut, long truncOut)
 {
-  long n, m;
-
   if (truncOut <= truncIn)
     {
-      for (n = 0; n <= truncOut; n++)
+      for (long n = 0; n <= truncOut; n++)
         {
-          for (m = n; m <= truncOut; m++)
+          for (long m = n; m <= truncOut; m++)
             {
               *arrayOut++ = *arrayIn++;
               *arrayOut++ = *arrayIn++;
@@ -535,21 +503,21 @@ sp2sp(double *arrayIn, long truncIn, double *arrayOut, long truncOut)
     }
   else
     {
-      for (n = 0; n <= truncIn; n++)
+      for (long n = 0; n <= truncIn; n++)
         {
-          for (m = n; m <= truncIn; m++)
+          for (long m = n; m <= truncIn; m++)
             {
               *arrayOut++ = *arrayIn++;
               *arrayOut++ = *arrayIn++;
             }
-          for (m = truncIn + 1; m <= truncOut; ++m)
+          for (long m = truncIn + 1; m <= truncOut; ++m)
             {
               *arrayOut++ = 0.0;
               *arrayOut++ = 0.0;
             }
         }
-      for (n = truncIn + 1; n <= truncOut; ++n)
-        for (m = n; m <= truncOut; ++m)
+      for (long n = truncIn + 1; n <= truncOut; ++n)
+        for (long m = n; m <= truncOut; ++m)
           {
             *arrayOut++ = 0.0;
             *arrayOut++ = 0.0;
@@ -564,11 +532,9 @@ sp2sp(double *arrayIn, long truncIn, double *arrayOut, long truncOut)
 void
 spcut(double *arrayIn, double *arrayOut, long trunc, const int *waves)
 {
-  long n, m;
-
-  for (n = 0; n <= trunc; n++)
+  for (long n = 0; n <= trunc; n++)
     {
-      for (m = n; m <= trunc; m++)
+      for (long m = n; m <= trunc; m++)
         {
           if (waves[m])
             {
